#'Compute the Standardised Precipitation-Evapotranspiration Index
#'
#'Calculation of the Standardised Precipitation-Evapotranspiration Index (SPEI) 
#'that is a multiscalar drought index based on climatic data. It can be used for 
#'determining the onset, duration and magnitude of drought conditions with 
#'respect to normal conditions in a variety of natural and managed systems such 
#'as crops, ecosystems, rivers, water resources, etc. The SPI is calculated 
#'using monthly (or weekly) precipitation as the input data. The SPEI uses the 
#'monthly (or weekly) difference between precipitation and pet. This represents 
#'a simple climatic water balance which is calculated at different time scales 
#'to obtain the SPEI.
#' 
#'@param exp A named list with the needed \code{s2dv_cube} objects containing 
#'  the seasonal forecast experiment in the data element for each variable. 
#'  Specific variables are needed for each method used in computing the 
#'  Potential Evapotranspiration. See parameter 'pet_method'. The accepted 
#'  variables are: 'tasmin' and 'tasmax', required for methods 'hargreaves' and 
#'  'hargreaves_modified' and 'tas', required for method 'thornthwaite'. 
#'  Variable 'prlr' is always needed. The units for temperature variables
#'  ('tasmin', 'tasmax' and 'tas') need to be in Celcius degrees; the units for 
#'  precipitation ('prlr') need to be in mm/month. 
#'@param exp_cor A named list with the needed \code{s2dv_cube} objects for each 
#'  variable in which the quantile PeriodSPEI should be applied. If it is not 
#'  specified, the PeriodSPEI is calculated from object 'exp'.
#'@param time_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'syear'. 
#'@param leadtime_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'time'. 
#'@param memb_dim A character string indicating the name of the dimension in 
#'  which the ensemble members are stored. When set it to NULL, threshold is 
#'  computed for individual members.
#'@param lat_dim A character string indicating the name of the latitudinal 
#'  dimension. By default it is set by 'latitude'.
#'@param accum An integer value indicating the number of months for the 
#'  accumulation for each variable.
#'@param ref_period A list with two numeric values with the starting and end 
#'  points of the reference period used for computing the index. The default 
#'  value is NULL indicating that the first and end values in data will be 
#'  used as starting and end points.
#'@param params A multidimensional array with named dimensions for computing the 
#'  SPEI. This option overrides computation of fitting parameters. It needs 
#'  to be of same leadtime and time dimensions of 'exp' and a dimension named 
#'  'coef' with the length of the coefficients needed for the used distribution 
#'  (for 'Gamma' coef dimension is of lenght 2, for 'log-Logistic' or 
#'  'PearsonIII' is of length 3). If cross_validation is FALSE, the leadtime 
#'  dimension must be of length 1.
#'@param pet_exp A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp'. It must have the same dimensions of the 
#'  variable 'prlr' of 'exp'. If it is NULL it is calculated using the provided 
#'  variables with specified 'pet_method'. It is NULL by default.
#'@param pet_expcor A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp_cor'. It must have the same dimensions of 
#'  the variable 'prlr' of 'exp_cor'. If it is NULL it is calculated using the 
#'  provided variables with specified 'pet_method'. It is NULL by default.
#'@param standardization A logical value indicating wether the standardization 
#'  is computed.
#'@param cross_validation A logical value indicating if cross validation is 
#'  done (TRUE), or not (FALSE). It is FALSE by default.
#'@param pet_method A character string indicating the method used to compute 
#'  the potential evapotranspiration. The accepted methods are:
#'  'hargreaves' and 'hargreaves_modified', that require the data to have 
#'  variables tasmin and tasmax; and 'thornthwaite', that requires variable 
#'  'tas'.  
#'@param method A character string indicating the standardization method used. 
#'  If can be: 'parametric' or 'non-parametric'. It is set to 'parametric' by 
#'  default.
#'@param distribution A character string indicating the name of the distribution 
#'  function to be used for computing the SPEI. The accepted names are: 
#'  'log-Logistic', 'Gamma' or 'PearsonIII'. It is set to 'log-Logistic' by 
#'  default.
#'@param fit A character string indicating the name of the method used for 
#'  computing the distribution function parameters The accepteed names are: 
#'  'ub-pwm', 'pp-pwm' and 'max-lik'. It is set to 'ub-pwm' by default.
#'@param param_error A numeric value with the error accepted.
#'@param handle_infinity A logical value wether to return Infinite values (TRUE)
#'  or not (FALSE).
#'@param na.rm A logical value indicating whether NA values should be removed 
#'  from data. It is FALSE by default.
#'@param ncores An integer value indicating the number of cores to use in 
#'  parallel computation.
#' 
#'@return A list with elements:
#'\itemize{
#'  \item{'exp', if 'standarization' is TRUE an 's2dv_cube' conaining the 'SPEI' 
#'        in element data from 'exp' array with the same dimensions as 'exp'.
#'        If it is FALSE, it is an array with the accumulated values of PET 
#'        minus 'prlr' data.}
#'  \item{'exp_cor', if 'standarization' is TRUE and if 'exp_cor' is not 
#'        NULL. It is an 's2dv_cube' with the SPEI data from 'exp_cor' in 
#'        element 'data'. If 'standarization' is FALSE, only the accumulated 
#'        values of PET minus 'prlr' is returned.}
#'  \item{'params', returned if 'standarization' is TRUE, it contains the 
#'        parameters used for the standarization of 'exp' that are used for 
#'        computing the 'SPEI' for 'exp_cor'.}
#' }
#' 
#'@examples 
#'dims <- c(var = 1, sday = 1, sweek = 1, syear = 6, time = 3, 
#'          latitude = 2, longitude = 1, ensemble = 25)
#'dimscor <- c(var = 1, sday = 1, sweek = 1, syear = 1, time = 3, 
#'             latitude = 2, longitude = 1, ensemble = 15)
#'
#'dates_exp <- as.POSIXct(c(paste0(2010:2015, "-08-16"), 
#'                          paste0(2010:2015, "-09-15"), 
#'                          paste0(2010:2015, "-10-16")), 'UTC')
#'dim(dates_exp) <- c(sday = 1, sweek = 1, syear = 6, time = 3)
#'dates_expcor <- as.POSIXct(c(paste0(2020, "-08-16"), paste0(2020, "-09-15"), 
#'                             paste0(2020, "-10-16")), 'UTC')
#'dim(dates_expcor) <- c(sday = 1, sweek = 1, syear = 1, time = 3)
#'lat <- c(40,40.1)
#'
#'exp_tasmax <- array(rnorm(900, 27.73, 5.26), dim = dims)
#'exp_tasmin <- array(rnorm(900, 14.83, 3.86), dim = dims)
#'exp_prlr <- array(rnorm(900, 21.19, 25.64), dim = dims)
#'
#'expcor_tasmax <- array(rnorm(540, 29.03, 5.67), dim = dimscor)
#'expcor_tasmin <- array(rnorm(540, 15.70, 4.40), dim = dimscor)
#'expcor_prlr <- array(rnorm(540, 15.62, 21.38), dim = dimscor)
#'
#'exp <- list('tasmax' = exp_tasmax, 'tasmin' = exp_tasmin, 'prlr' = exp_prlr)
#'exp_cor <- list('tasmax' = expcor_tasmax, 'tasmin' = expcor_tasmin, 
#'                'prlr' = expcor_prlr)
#'
#'exp <- lapply(exp, CSTools::s2dv_cube, coords = list(latitude = lat), 
#'              Dates = dates_exp)
#'exp_cor <- lapply(exp_cor, CSTools::s2dv_cube, coords = list(latitude = lat), 
#'                  Dates = dates_expcor)
#'
#'res <- CST_PeriodSPEI(exp = exp, exp_cor = exp_cor)
#' 
#'@import multiApply
#'@import ClimProjDiags
#'@import SPEI
#'@import zoo
#'@import TLMoments
#'@import lmomco
#'@import lmom
#'@import lubridate
#'@import CSTools
#'@export
CST_PeriodSPEI <- function(exp, exp_cor = NULL, 
                           time_dim = 'syear', leadtime_dim = 'time', 
                           memb_dim =  'ensemble', lat_dim = 'latitude',
                           accum = 1, ref_period = NULL, params = NULL,
                           pet_exp = NULL, pet_expcor = NULL, 
                           standardization = TRUE, cross_validation = FALSE, 
                           pet_method = 'hargreaves', method = 'parametric', 
                           distribution = 'log-Logistic', fit = 'ub-pwm', 
                           param_error = -9999, handle_infinity = FALSE, 
                           na.rm = FALSE, ncores = NULL) {
                          
  # Check 's2dv_cube'
  if (is.null(exp)) {
    stop("Parameter 'exp' cannot be NULL.")
  }
  if (!all(sapply(exp, function(x) inherits(x, 's2dv_cube')))) {
    stop("Parameter 'exp' must be a list of 's2dv_cube' class.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) inherits(x, 's2dv_cube')))) {
      stop("Parameter 'exp_cor' must be a list of 's2dv_cube' class.")
    }
  }
  # latitude
  if (!any(names(exp[[1]]$coords) %in% .KnownLatNames())) {
    stop("Spatial coordinate names of parameter 'exp' do not match any ",
         "of the names accepted by the package.")
  }
  # Dates
  dates_exp <- exp[[1]]$attrs$Dates
  if (!'Dates' %in% names(exp[[1]]$attrs)) {
    stop("Element 'Dates' is not found in 'attrs' list of 'exp'. ", 
         "See 's2dv_cube' object description in README file for more ", 
         "information.")
  }

  if (!is.null(exp_cor)) {
    if (!'Dates' %in% names(exp_cor[[1]]$attrs)) {
      stop("Element 'Dates' is not found in 'attrs' list of 'exp_cor'. ", 
           "See 's2dv_cube' object description in README file for more ", 
           "information.")
    }
  }

  lat_dim <- names(exp[[1]]$coords)[[which(names(exp[[1]]$coords) %in% .KnownLatNames())]]

  res <- PeriodSPEI(exp = lapply(exp, function(x) x$data), 
                    dates_exp = exp[[1]]$attrs$Dates, 
                    lat = exp[[1]]$coords[[lat_dim]], 
                    exp_cor = if (is.null(exp_cor)) {NULL} else {
                                  lapply(exp_cor, function(x) x$data)}, 
                    dates_expcor = exp_cor[[1]]$attrs$Dates, 
                    time_dim = time_dim, leadtime_dim = leadtime_dim, 
                    memb_dim = memb_dim, lat_dim = lat_dim, 
                    accum = accum, ref_period = ref_period, params = params, 
                    pet_exp = pet_exp, pet_expcor = pet_expcor, 
                    standardization = standardization, 
                    cross_validation = cross_validation, 
                    pet_method = pet_method, method = method, 
                    distribution = distribution, fit = fit, 
                    param_error = param_error, 
                    handle_infinity = handle_infinity, na.rm = na.rm,
                    ncores = ncores)
  if (is.null(exp_cor)) {
    source_files <- lapply(exp, function(x) {x$attrs$source_files})
    suppressWarnings(
      res_exp <- CSTools::s2dv_cube(data = res[[1]], coords = exp[[1]]$coords, 
                                    varName = c('SPEI'), Dates = dates_exp, 
                                    when = Sys.time(), 
                                    source_files = source_files)
    )
    if (standardization) {
      return(list(spei_exp = res_exp, params = res[[2]]))
    } else {
      return(exp)
    }
  } else {
    source_files_exp <- lapply(exp, function(x) {x$attrs$source_files})
    suppressWarnings(
      res_exp <- CSTools::s2dv_cube(data = res[[1]], coords = exp[[1]]$coords, 
                                    varName = c('SPEI'), Dates = dates_exp, 
                                    source_files = source_files_exp,
                                    when = Sys.time())
    )
    source_files_expcor <- lapply(exp_cor, function(x) {x$attrs$source_files})
    suppressWarnings(
      res_expcor <- CSTools::s2dv_cube(data = res[[2]], coords = exp_cor[[1]]$coords, 
                                      varName = c('SPEI'), Dates = dates_expcor, 
                                      source_files = source_files_expcor, 
                                      when = Sys.time())
    )

    if (standardization) {
      return(list(spei_exp = res_exp, spei_exp_cor = res_expcor, params = res[[3]]))
    } else {
      return(list(spei_exp = res_exp, spei_exp_cor = res_expcor))
    }
  }
  return(res)
}


#'Compute the Standardised Precipitation-Evapotranspiration Index
#'
#'Calculation of the Standardised Precipitation-Evapotranspiration Index (SPEI) 
#'that is a multiscalar drought index based on climatic data. It can be used for 
#'determining the onset, duration and magnitude of drought conditions with 
#'respect to normal conditions in a variety of natural and managed systems such 
#'as crops, ecosystems, rivers, water resources, etc. The SPI is calculated 
#'using monthly (or weekly) precipitation as the input data. The SPEI uses the 
#'monthly (or weekly) difference between precipitation and pet. This represents 
#'a simple climatic water balance which is calculated at different time scales 
#'to obtain the SPEI.
#' 
#'@param exp A named list with the needed \code{s2dv_cube} objects containing 
#'  the seasonal forecast experiment in the data element for each variable. 
#'  Specific variables are needed for each method used in computing the 
#'  Potential Evapotranspiration. See parameter 'pet_method'. The accepted 
#'  variables are: 'tasmin' and 'tasmax', required for methods 'hargreaves' and 
#'  'hargreaves_modified' and 'tas', required for method 'thornthwaite'. 
#'  Variable 'prlr' is always needed. The units for temperature variables
#'  ('tasmin', 'tasmax' and 'tas') need to be in Celcius degrees; the units for 
#'  precipitation ('prlr') need to be in mm/month. 
#'@param dates_exp An array of temporal dimensions containing the Dates of 
#'  'exp'. It must be of class 'Date' or 'POSIXct'.
#'@param lat A numeric vector containing the latitude values of 'exp'.
#'@param exp_cor A named list with the needed \code{s2dv_cube} objects for each 
#'  variable in which the quantile PeriodSPEI should be applied. If it is not 
#'  specified, the PeriodSPEI is calculated from object 'exp'.
#'@param dates_expcor An array of temporal dimensions containing the Dates of 
#'  'exp_cor'. It must be of class 'Date' or 'POSIXct'.
#'@param time_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'syear'. 
#'@param leadtime_dim A character string indicating the name of the temporal 
#'  dimension. By default, it is set to 'time'. 
#'@param memb_dim A character string indicating the name of the dimension in 
#'  which the ensemble members are stored. When set it to NULL, threshold is 
#'  computed for individual members.
#'@param lat_dim A character string indicating the name of the latitudinal 
#'  dimension. By default it is set by 'latitude'.
#'@param accum An integer value indicating the number of months for the 
#'  accumulation for each variable.
#'@param ref_period A list with two numeric values with the starting and end 
#'  points of the reference period used for computing the index. The default 
#'  value is NULL indicating that the first and end values in data will be 
#'  used as starting and end points.
#'@param params A multidimensional array with named dimensions for computing the 
#'  SPEI. This option overrides computation of fitting parameters. It needs 
#'  to be of same leadtime and time dimensions of exp and a dimension named 
#'  'coef' with the length of the coefficients needed for the used distribution 
#'  (for 'Gamma' coef dimension is of lenght 2, for 'log-Logistic' or 
#'  'PearsonIII' is of length 3). If cross_validation is FALSE, the leadtime 
#'  dimension must be of length 1.
#'@param pet_exp A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp'. It must have the same dimensions of the 
#'  variable 'prlr' of 'exp'. If it is NULL it is calculated using the provided 
#'  variables with specified 'pet_method'. It is NULL by default.
#'@param pet_expcor A multidimensional array containing the Potential  
#'  EvapoTranspiration data of 'exp_cor'. It must have the same dimensions of 
#'  the variable 'prlr' of 'exp_cor'. If it is NULL it is calculated using the 
#'  provided variables with specified 'pet_method'. It is NULL by default.
#'@param standardization A logical value indicating wether the standardization 
#'  is computed.
#'@param cross_validation A logical value indicating if cross validation is 
#'  done (TRUE), or not (FALSE). It is FALSE by default.
#'@param pet_method A character string indicating the method used to compute 
#'  the potential evapotranspiration. The accepted methods are:
#'  'hargreaves' and 'hargreaves_modified', that require the data to have 
#'  variables tasmin and tasmax; and 'thornthwaite', that requires variable 
#'  'tas'.  
#'@param method A character string indicating the standardization method used. 
#'  If can be: 'parametric' or 'non-parametric'.
#'@param distribution A character string indicating the name of the distribution 
#'  function to be used for computing the SPEI. The accepted names are: 
#'  'log-Logistic', 'Gamma' or 'PearsonIII'. It is set to 'log-Logistic' by 
#'  default.
#'@param fit A character string indicating the name of the method used for 
#'  computing the distribution function parameters The accepteed names are: 
#'  'ub-pwm', 'pp-pwm' and 'max-lik'. It is set to 'ub-pwm' by default.
#'@param param_error A numeric value with the error accepted.
#'@param handle_infinity A logical value wether to return Infinite values (TRUE)
#'  or not (FALSE).
#'@param na.rm A logical value indicating whether NA values should be removed 
#'  from data. It is FALSE by default.
#'@param ncores An integer value indicating the number of cores to use in 
#'  parallel computation.
#' 
#'@return A list with elements:
#'\itemize{
#'  \item{'exp', if 'standarization' is TRUE an array conaining SPEI data from 
#'        'exp' array with the same dimensions as 'exp'. If it is FALSE, it 
#'         is an array with the accumulated values of PET minus 'prlr' data.}
#'  \item{'exp_cor', if 'standarization' is TRUE and if 'exp_cor' is not 
#'        NULL. It is an array with the SPEI data from 'exp_cor'. If 
#'        'standarization' is FALSE, only the accumulated values of PET minus 
#'        'prlr' is returned.}
#'  \item{'params', returned if 'standarization' is TRUE, it contains the 
#'        parameters used for the standarization of 'exp' that are used for 
#'        computing the 'SPEI' for 'exp_cor'.}
#'}
#' 
#'@examples 
#'dims <- c(var = 1, sday = 1, sweek = 1, syear = 6, time = 3, 
#'          latitude = 2, longitude = 1, ensemble = 25)
#'dimscor <- c(var = 1, sday = 1, sweek = 1, syear = 1, time = 3, 
#'             latitude = 2, longitude = 1, ensemble = 15)
#'exp_tasmax <- array(rnorm(900, 27.73, 5.26), dim = dims)
#'exp_tasmin <- array(rnorm(900, 14.83, 3.86), dim = dims)
#'exp_prlr <- array(rnorm(900, 21.19, 25.64), dim = dims)
#' 
#'expcor_tasmax <- array(rnorm(540, 29.03, 5.67), dim = dimscor)
#'expcor_tasmin <- array(rnorm(540, 15.70, 4.40), dim = dimscor)
#'expcor_prlr <- array(rnorm(540, 15.62, 21.38), dim = dimscor)
#' 
#'dates_exp <- as.Date(c(paste0(2010:2015, "-08-16"), 
#'                       paste0(2010:2015, "-09-15"), 
#'                       paste0(2010:2015, "-10-16")))
#'dim(dates_exp) <- c(sday = 1, sweek = 1, syear = 6, time = 3)
#'dates_expcor <- as.POSIXct(c(paste0(2020, "-08-16"), 
#'                             paste0(2020, "-09-15"), 
#'                             paste0(2020, "-10-16")), 'UTC')
#'dim(dates_expcor) <- c(sday = 1, sweek = 1, syear = 1, time = 3)
#'lat <- c(40,40.1)
#' 
#'exp <- list('tasmax' = exp_tasmax, 'tasmin' = exp_tasmin, 'prlr' = exp_prlr)
#'exp_cor <- list('tasmax' = expcor_tasmax, 'tasmin' = expcor_tasmin, 
#'                'prlr' = expcor_prlr)
#'res <- PeriodSPEI(exp = exp, exp_cor = exp_cor, lat = lat, 
#'                  dates_exp = dates_exp, dates_expcor = dates_expcor)
#' 
#'@import multiApply
#'@import ClimProjDiags
#'@import SPEI
#'@import zoo
#'@import TLMoments
#'@import lmomco
#'@import lmom
#'@import lubridate
#'@export
PeriodSPEI <- function(exp, dates_exp, lat, 
                       exp_cor = NULL, dates_expcor = NULL,
                       time_dim = 'syear', leadtime_dim = 'time', 
                       memb_dim = 'ensemble', lat_dim = 'latitude',
                       accum = 1, ref_period = NULL, params = NULL, 
                       pet_exp = NULL, pet_expcor = NULL, 
                       standardization = TRUE, cross_validation = FALSE,
                       pet_method = 'hargreaves', method = 'parametric', 
                       distribution = 'log-Logistic', fit = 'ub-pwm', 
                       param_error = -9999, handle_infinity = FALSE, 
                       na.rm = FALSE, ncores = NULL) {

  # Initial checks
  ## exp
  if (!inherits(exp, 'list')) {
    stop("Parameter 'exp' needs to be a named list with the needed variables.")
  }
  if (is.null(names(exp))) {
    stop("Parameter 'exp' needs to be a named list with the variable names.")
  }
  if (any(sapply(exp, function(x) is.null(names(dim(x)))))) {
    stop("Parameter 'exp' needs to be a list of arrays with dimension names.")
  }
  dims <- lapply(exp, function(x) dim(x))
  first_dims <- dims[[1]]
  all_equal <- all(sapply(dims[-1], function(x) identical(first_dims, x)))
  if (!all_equal) {
    stop("Parameter 'exp' variables needs to have the same dimension names.")
  }

  ## exp_cor
  if (!is.null(exp_cor)) {
    if (!inherits(exp_cor, 'list')) {
      stop("Parameter 'exp_cor' needs to be a named list with the needed ", 
           "variables if it is not NULL.")
    }
    if (is.null(names(exp_cor))) {
      stop("Parameter 'exp_cor' needs to be a named list with the variable names.")
    }
    if (any(sapply(exp_cor, function(x) is.null(names(dim(x)))))) {
      stop("Parameter 'exp_cor' needs to be a list of arrays with dimension names.")
    }
    dimscor <- lapply(exp_cor, function(x) dim(x))
    first_dimscor <- dimscor[[1]]
    all_equal <- all(sapply(dimscor[-1], function(x) identical(first_dimscor, x)))
    if (!all_equal) {
      stop("Parameter 'exp_cor' variables needs to have the same dimension names.")
    }
  }

  # lat
  if (!is.numeric(lat)) {
    stop("Parameter 'lat' must be numeric.")
  }
  if (any(sapply(dims, FUN = function(x) x[lat_dim] != length(lat)))) {
    stop("Parameter 'lat' needs to have the same length of latitudinal", 
         "dimension of all the variables arrays in 'exp'.")
  }
  if (!is.null(exp_cor)) {
    if (any(sapply(dimscor, FUN = function(x) x[lat_dim] != length(lat)))) {
      stop("Parameter 'lat' needs to have the same length of latitudinal", 
          "dimension of all the variables arrays in 'exp'.")
    }
  }

  # Variable checks
  ## exp (2)
  pet <- vector("list", 2)
  if (is.null(pet_exp)) {
    if (all(c('tasmin', 'tasmax', 'prlr') %in% names(exp))) {
      # hargreaves modified: 'tasmin', 'tasmax', 'prlr' and 'lat'
      if (!(pet_method[1] %in% c('hargreaves_modified', 'hargreaves'))) {
        warning("Parameter 'pet_method' needs to be 'hargreaves' or ", 
                "'hargreaves_modified'. It is set to 'hargreaves'.")
        pet_method[1] <- 'hargreaves'
      }
    } else if (all(c('tas', 'prlr') %in% names(exp))) {
      # thornthwaite: 'tas' (mean), 'lat'
      if (!(pet_method[1] == 'thornthwaite')) {
        warning("Parameter 'pet_method' it is set to be 'thornthwaite'.")
        pet_method[1] <- 'thornthwaite'
      }
    } else {
      stop("Parameter 'exp' needs to be a named list with accepted ", 
           "variable names if 'pet_exp' is not provided. See documentation.")
    }
  } else {
    if (!('prlr' %in% names(exp))) {
      stop("Variable 'prlr' is not included in 'exp'.")
    }
    if (length(dim(exp[['prlr']])) != length(dim(pet_exp))) {
      stop("Parameter 'pet_exp' must have the same length of all the ", 
            "dimensions as variable 'prlr' in 'exp'.") 
    }
    if (!all(dim(exp[['prlr']]) %in% dim(pet_exp))) {
      stop("Parameter 'pet_exp' must have the same length of all the ", 
           "dimensions as variable 'prlr' in 'exp'.")
    }
    if (any(names(dim(exp[['prlr']])) != names(dim(pet_exp)))) {
      pos <- match(names(dim(exp[['prlr']])), names(dim(pet_exp)))
      pet_exp <- aperm(pet_exp, pos)
    }
    pet[[1]] <- pet_exp
  }
  ## exp_cor (2)
  if (!is.null(exp_cor)) {
    if (is.null(pet_expcor)) {
      if (length(exp_cor) < 1) {
        exp_cor <- NULL
      } else {
        if (length(pet_method) == 1) {
          pet_method <- rep(pet_method, 2)
        }
      }
      if (all(c('tasmin', 'tasmax', 'prlr') %in% names(exp_cor))) {
        # hargreaves modified: 'tasmin', 'tasmax', 'prlr' and 'lat'
        if (!(pet_method[2] %in% c('hargreaves_modified', 'hargreaves'))) {
          warning("Parameter 'pet_method' needs to be 'hargreaves' or ", 
                  "'hargreaves_modified'. It is set to 'hargreaves'.")
          pet_method[2] <- 'hargreaves'
        }
      } else if (all(c('tas', 'prlr') %in% names(exp_cor))) {
        # thornthwaite: 'tas' (mean), 'lat'
        if (!(pet_method[2] == 'thornthwaite')) {
          warning("Parameter 'pet_method' it is set to be 'thornthwaite'.")
          pet_method[2] <- 'thornthwaite'
        }
      } else {
        stop("Parameter 'exp_cor' needs to be a list with the needed ", 
             "variables if 'pet_expcor' is not provided.")
      }
    } else {
      if (!is.null(exp_cor)) {
        if (!('prlr' %in% names(exp_cor))) {
          stop("Variable 'prlr' is not included in 'exp_cor'.")
        }
        if (length(dim(exp_cor[['prlr']])) != length(dim(pet_expcor))) {
          stop("Parameter 'pet_expcor' must have the same length of all the ", 
                "dimensions as variable 'prlr' in 'exp_cor'.") 
        }
        if (!all(dim(exp_cor[['prlr']]) %in% dim(pet_expcor))) {
          stop("Parameter 'pet_expcor' must have the same length of all the ", 
              "dimensions as variable 'prlr' in 'exp_cor'.")
        }
        if (any(names(dim(exp_cor[['prlr']])) != names(dim(pet_expcor)))) {
          pos <- match(names(dim(exp_cor[['prlr']])), names(dim(pet_expcor)))
          pet_expcor <- aperm(pet_expcor, pos)
        }
      }
      pet[[2]] <- pet_expcor
    }
  }

  ## time_dim
  if (!is.character(time_dim) | length(time_dim) != 1) {
    stop("Parameter 'time_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) time_dim %in% names(dim(x))))) {
    stop("Parameter 'time_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) time_dim %in% names(dim(x))))) {
      stop("Parameter 'time_dim' is not found in 'exp_cor' dimension.")
    }
  }
  ## leadtime_dim
  if (!is.character(leadtime_dim) | length(leadtime_dim) != 1) {
    stop("Parameter 'leadtime_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) leadtime_dim %in% names(dim(x))))) {
    stop("Parameter 'leadtime_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) leadtime_dim %in% names(dim(x))))) {
      stop("Parameter 'leadtime_dim' is not found in 'exp_cor' dimension.")
    }
  }
  ## memb_dim
  if (!is.character(memb_dim) | length(memb_dim) != 1) {
    stop("Parameter 'time_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) memb_dim %in% names(dim(x))))) {
    stop("Parameter 'memb_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) memb_dim %in% names(dim(x))))) {
      stop("Parameter 'memb_dim' is not found in 'exp_cor' dimension.")
    }
  }
  ## lat_dim
  if (!is.character(lat_dim) | length(lat_dim) != 1) {
    stop("Parameter 'lat_dim' must be a character string.")
  }
  if (!all(sapply(exp, function(x) lat_dim %in% names(dim(x))))) {
    stop("Parameter 'lat_dim' is not found in 'exp' dimension.")
  }
  if (!is.null(exp_cor)) {
    if (!all(sapply(exp_cor, function(x) lat_dim %in% names(dim(x))))) {
      stop("Parameter 'lat_dim' is not found in 'exp_cor' dimension.")
    }
  }
  # dates
  if (is.null(dates_exp)) {
    stop("Parameter 'dates_exp' is missing, dates must be provided.")
  }
  if (!(is.Date(dates_exp)) & !(is.POSIXct(dates_exp))) {
    stop("Parameter 'dates_exp' is not of the correct class, ", 
         "only 'Date' and 'POSIXct' classes are accepted.")
  }
  if (!all(dim(exp[[1]])[c(time_dim, leadtime_dim)] == 
           dim(dates_exp)[c(time_dim, leadtime_dim)])) {
    stop("Parameter 'dates_exp' needs to have the same length as 'time_dim' ", 
         "and 'leadtime_dim' as 'exp'.")
  }

  if (!is.null(exp_cor)) {
    if (is.null(dates_expcor)) {
      stop("Parameter 'dates_expcor' is missing, dates for 'exp_cor' must be ", 
           "provided if exp_cor is not NULL.")
    }
    if (!(is.Date(dates_expcor)) & !(is.POSIXct(dates_expcor))) {
      stop("Element 'Dates' in 'exp_cor' is not of the correct class, ", 
           "only 'Date' and 'POSIXct' classes are accepted.")
    }
    if (!all(dim(exp_cor[[1]])[c(time_dim, leadtime_dim)] == 
             dim(dates_expcor)[c(time_dim, leadtime_dim)])) {
      stop("Parameter 'dates_expcor' needs to have the same length as ", 
           "'time_dim' and 'leadtime_dim' as 'exp_cor'.")
    }
  }
  ## accum
  if (accum > dim(exp[[1]])[leadtime_dim]) {
    stop(paste0("Cannot compute accumulation of ", accum, " months because ", 
                "loaded data has only ", dim(exp[[1]])[leadtime_dim], " months."))
  }
  ## params
  if (!is.null(params)) {
    if (!is.numeric(params)) {
      stop("Parameter 'params' must be numeric.")
    }
    if (!all(c(time_dim, leadtime_dim, 'coef') %in% names(dim(params)))) {
      stop("Parameter 'params' must be a multidimensional array with named ", 
           "dimensions: 'time_dim', 'leadtime_dim' and 'coef'.")
    }
  }

  ## standardization
  if (!is.logical(standardization)) {
    stop("Parameter 'standardization' must be a logical value.")
  }

  ## param_error
  if (!is.numeric(param_error)) {
    stop("Parameter 'param_error' must be a numeric value.")
  }

  ## handle_infinity
  if (!is.logical(handle_infinity)) {
    stop("Parameter 'handle_infinity' must be a logical value.")
  }

  ## cross_validation
  if (!is.logical(cross_validation)) {
    stop("Parameter 'cross_validation' must be a logical value.")
  }
  # if (cross_validation) {
  #   warning("Detected 'cross_validation' = TRUE. This functionality ", 
  #           "is being developed, sorry for the inconvenience. It will ", 
  #           "be set to FALSE.")
  #   cross_validation <- FALSE
  # }

  ## method
  if (!(method %in% c('parametric', 'non-parametric'))) {
    stop("Parameter 'method' must be a character string containing one of ", 
         "the following methods: 'parametric' or 'non-parametric'.")
  }
  ## distribution
  if (!(distribution %in% c('log-Logistic', 'Gamma', 'PearsonIII'))) {
    stop("Parameter 'distribution' must be a character string containing one ", 
         "of the following distributions: 'log-Logistic', 'Gamma' or ", 
         "'PearsonIII'.")
  }
  ## fit
  if (!(fit %in% c('max-lik', 'ub-pwm', 'pp-pwm'))) {
    stop("Parameter 'fit' must be a character string containing one of ", 
         "the following fit methods: 'max-lik', 'ub-pwm', 'pp-pwm'.")
  }
  ## ncores
  if (!is.null(ncores)) {
    if (!is.numeric(ncores) | any(ncores %% 1 != 0) | any(ncores < 0) |
        length(ncores) > 1) {
      stop("Parameter 'ncores' must be a positive integer.")
    }
  }
  # Data preparation
  
  # complete dates
  dates_monthly <- NULL
  k = 0
  for (dates in .return2list(dates_exp, dates_expcor)) {
    k = k + 1
    ini_date <- as.Date(paste(lubridate::year(min(dates)), 01, 01, sep= '-'))
    end_date <- as.Date(paste(lubridate::year(max(dates)), 12, 31, sep='-'))
    dates_complete_daily <- as.Date(ini_date:end_date)

    dates_complete_daily_to_monthly <- lubridate::day(dates_complete_daily)
    dates_complete_monthly <- dates_complete_daily[which(dates_complete_daily_to_monthly == 1)]
    dates_monthly[[k]] <- array(0, dim = length(dates_complete_monthly))
    for (dd in 1:length(dates)) {
      ii <- which(dates_complete_monthly == as.Date(paste(lubridate::year(dates[dd]), 
                                                          lubridate::month(dates[dd]), 
                                                          01, sep = '-')))
      dates_monthly[[k]][ii] <- 1
    }
  }

  # Compute PeriodSPEI
  k = 0
  spei_res <- NULL
  computed_pet <- FALSE

  for (data in .return2list(exp, exp_cor)) {
    k = k + 1
    # Evapotranspiration estimation (unless pet is already provided)
    if (is.null(pet[[k]]) | computed_pet) {
      pet[[k]] <- .Evapotranspiration(data = data, 
                                      dates_monthly = dates_monthly[[k]], 
                                      lat = lat, pet_method = pet_method[k], 
                                      time_dim = time_dim, 
                                      leadtime_dim = leadtime_dim, 
                                      lat_dim = lat_dim, na.rm = na.rm, 
                                      ncores = ncores)
      computed_pet <- TRUE
    }

    # Accumulation
    diff_p_pet <- data$prlr - pet[[k]]
    data_accum <- .Accumulation(data = diff_p_pet, 
                                dates_monthly = dates_monthly[[k]], accum = accum, 
                                time_dim = time_dim, leadtime_dim = leadtime_dim, 
                                ncores = ncores)
    # Standardization
    if (standardization) {
      spei_dat <- .Standardization(data = data_accum, params = params, 
                                   accum = accum, time_dim = time_dim,
                                   leadtime_dim = leadtime_dim,
                                   memb_dim = memb_dim, 
                                   ref_period = ref_period,
                                   cross_validation = cross_validation,
                                   handle_infinity = handle_infinity,
                                   param_error = param_error,
                                   method = method, distribution = distribution, 
                                   fit = fit, ncores = ncores)
      params <- spei_dat$params

      pos <- match(names(dim(data[[1]])), names(dim(spei_dat[[1]])))
      spei_dat[[1]] <- aperm(spei_dat[[1]], pos)

      spei_res[[k]] <- spei_dat[[1]]
    } else {

      pos <- match(names(dim(data[[1]])), names(dim(data_accum)))
      data_accum <- aperm(data_accum, pos)

      spei_res[[k]] <- data_accum
    }
  }

  if (standardization) {
    spei_res[[k+1]] <- params
    if (is.null(exp_cor)) {
      names(spei_res) <- c('exp', 'params')
    } else {
      names(spei_res) <- c('exp', 'exp_cor', 'params')
    }
  } else {
    if (is.null(exp_cor)) {
      names(spei_res) <- c('exp')
    } else {
      names(spei_res) <- c('exp', 'exp_cor')
    }
  }
  return(spei_res)
}


.return2list <- function(data1, data2 = NULL) {
  if (is.null(data1) & is.null(data2)) {
    return(NULL)
  } else if (is.null(data2)) {
    return(list(data1))
  } else {
    return(list(data1, data2))
  }
}


.Evapotranspiration <- function(data, dates_monthly, lat, pet_method = 'hargreaves', 
                                time_dim = 'syear', leadtime_dim = 'time',
                                lat_dim = 'latitude', na.rm = FALSE, 
                                ncores = NULL) {

  lat_mask <- array(lat, dim = c(1, length(lat)))
  names(dim(lat_mask)) <- c('dat', lat_dim)

  # extract mask of NA locations to return to NA the final result
  mask_na <- array(1, dim = dim(data[[1]]))
  if (pet_method == 'hargreaves') {
    varnames <- c('tasmax', 'tasmin')
    mask_na[which(is.na(data$tasmax))] <- 0
    mask_na[which(is.na(data$tasmin))] <- 0
  } else if (pet_method == 'hargreaves_modified') {
    varnames <- c('tasmax', 'tasmin', 'prlr')
    mask_na[which(is.na(data$tasmax))] <- 0
    mask_na[which(is.na(data$tasmin))] <- 0
    mask_na[which(is.na(data$prlr))] <- 0
  } else if (pet_method == 'thornthwaite') {
    varnames <- c('tas')
    mask_na[which(is.na(data$tas))] <- 0
  }

  # replace NA with 0
  for (dd in 1:length(data)) {
    data[[dd]][which(is.na(data[[dd]]))] <- 0
  }

  # prepare data
  target_dims_data <- lapply(data[varnames], function(x) rep(c(leadtime_dim, time_dim), 1))
  pet <- Apply(data = c(list(lat_mask = lat_mask), data[varnames]), 
               target_dims = c(list(lat_mask = 'dat'), target_dims_data), 
               fun = .evapotranspiration, 
               dates_monthly = dates_monthly, pet_method = pet_method, 
               leadtime_dim = leadtime_dim, time_dim = time_dim,
               output_dims = c(leadtime_dim, time_dim),
               ncores = ncores)$output1
  # reorder dims in pet_estimated
  pos <- match(names(dim(data[[1]])), names(dim(pet)))
  pet <- aperm(pet, pos)

  # restore original NAs from mask_na
  pet[which(mask_na == 0 )] <- NA

  return(pet)
}

.evapotranspiration <- function(lat_mask, data2, data3 = NULL, data4 = NULL, 
                                dates_monthly, pet_method = 'hargreaves', 
                                leadtime_dim = 'time', time_dim = 'syear') {
 
  dims <- dim(data2)

  # create a vector from data but adding 0 to achive complete time series 
  # of the considered period 
  # (starting in January of the first year) so that the solar radiation 
  # estimation is computed in each case for the correct month

  if (!is.null(data2)) {
    data_tmp <- as.vector(data2)
    data2 <- array(0, dim = length(dates_monthly))
    count <- 1
    for (dd in 1:length(dates_monthly)) {
      if (dates_monthly[dd] == 1) {
        data2[dd] <- data_tmp[count]
        count <- count + 1
      }
    }
    rm(data_tmp)
  }
  any(is.na(data3))
  if (!is.null(data3)) {
    data_tmp <- as.vector(data3)
    data3 <- array(0, dim = length(dates_monthly))
    count <- 1
    for (dd in 1:length(dates_monthly)) {
      if (dates_monthly[dd] == 1) {
        data3[dd] <- data_tmp[count]
        count <- count + 1
      }
    }
    rm(data_tmp)
  }
  if (!is.null(data4)) {
    data_tmp <- as.vector(data4)
    data4 <- array(0, dim = length(dates_monthly))
    count <- 1
    for (dd in 1:length(dates_monthly)) {
      if (dates_monthly[dd] == 1) {
        data4[dd] <- data_tmp[count]
        count <- count + 1
      }
    }
    rm(data_tmp)
  }
  if (pet_method == 'hargreaves') {
    pet <- hargreaves(Tmin = as.vector(data3), Tmax = as.vector(data2), 
                      lat = lat_mask, na.rm = na.rm, verbose = FALSE)
    # line to return the vector to the size of the actual original data
    pet <- pet[which(dates_monthly == 1)]
    pet <- array(pet, dim = dims)
  }

  if (pet_method == 'hargreaves_modified') {
    pet <- hargreaves(Tmin = as.vector(data3), Tmax = as.vector(data2), 
                      lat = lat_mask, Pre = as.vector(data4), na.rm = na.rm, 
                      verbose = FALSE)
    pet <- pet[which(dates_monthly == 1)]  
    pet <- array(pet, dim = dims)
  }
  
  if (pet_method == 'thornthwaite') {
    pet <- thornthwaite(as.vector(data2), lat = lat_mask, na.rm = na.rm, 
                        verbose = FALSE)
    # line to return the vector to the size of the actual original data
    pet <- pet[which(dates_monthly == 1)] 
    pet <- array(pet, dim = dims)
  }
  return(pet)
}


.Accumulation <- function(data, dates_monthly, accum = 1, 
                          time_dim = 'syear', leadtime_dim = 'time', 
                          ncores = NULL) {
  
  accumulated <- Apply(data = list(data), 
                       target_dims = list(data = c(leadtime_dim, time_dim)), 
                       dates_monthly = dates_monthly,
                       accum = accum,
                       output_dims = c(leadtime_dim, time_dim), 
                       leadtime_dim = leadtime_dim, time_dim = time_dim,
                       fun = .accumulation, 
                       ncores = ncores)$output1

  pos <- match(names(dim(accumulated)), names(dim(data)))
  accumulated <- aperm(accumulated, pos)

  return(accumulated)

}

.accumulation <- function(data, dates_monthly, accum = 1, 
                          time_dim = 'syear', leadtime_dim = 'time') {

  # data:[time, syear]
  dims <- dim(data)
  
  data_vector <- array(0, dim = length(dates_monthly))
  count <- 1
  for (dd in 1:length(dates_monthly)) {
    if (dates_monthly[dd] == 1) {
      data_vector[dd] <- as.vector(data)[count]
      count <- count + 1
    }
  }
  # Accumulation at different timescales
  # rollapply {zoo} A generic function for applying a function to rolling 
  # margins of an array.
  data_sum_x <- rollapply(data_vector, accum, sum) 
  # adds as many NAs as needed at the begining to account for the months that 
  # cannot be added (depends on accu) and so that the position in the vector 
  # corresponds to the accumulated of the previous months (instead of the 
  # accumulated of the next months)
  data_sum_x <- c(rep(NA, accum-1), data_sum_x) 
  # discard the months that don't appear in the original data
  data_sum_x <- data_sum_x[which(dates_monthly == 1)] 
  accum_result <- array(data_sum_x, dim = c(dims))
  # replace by NA when the accumulation corresponds to months that where not 
  # present in the original data
  if (accum > 1) {
    accum_result[1:(accum-1), ] <- NA 
  }
  return(accum_result)
}

.Standardization <- function(data, params = NULL, accum = 1, time_dim = 'syear', 
                             leadtime_dim = 'time', memb_dim = 'ensemble', 
                             ref_period = NULL, cross_validation = FALSE,  
                             handle_infinity = FALSE, param_error = -9999, 
                             method = 'parametric', distribution = 'log-Logistic', 
                             fit = 'ub-pwm', na.rm = FALSE, ncores = NULL) {

  nleadtime <- dim(data)[leadtime_dim]
  ntime <- dim(data)[time_dim]

  if (!cross_validation) {
    ntime <- 1
  }

  coef = switch(distribution,
                "Gamma" = array(NA, dim = 2, dimnames = list(c('alpha', 'beta'))),
                "log-Logistic" = array(NA, dim = 3, dimnames = list(c('xi', 'alpha', 'kappa'))),
                "PearsonIII" = array(NA, dim = 3, dimnames = list(c('mu', 'sigma', 'gamma'))))

  if (is.null(params)) {
    params <- array(NA, dim = c(ntime, nleadtime, length(coef)))
    names(dim(params)) <- c(time_dim, leadtime_dim, 'coef')
  } else if (length(dim(params)) < 2) {
    params <- array(params, dim = c(length(params), ntime, nleadtime))
    # dim(params): [time_dim, leadtime_dim, coef] 
    # with the values repeated each time_dim and leadtime_dim
    params <- aperm(params, c(2,3,1)) 
    names(dim(params)) <- c(time_dim, leadtime_dim, 'coef')
  } else {
    if (dim(params)['coef'] != length(coef)) {
      stop(paste0("Params array should have 'coef' dimension of length ", 
                   length(coef), "."))
    }
  }

  spei <- Apply(data = list(data = data, params = params), 
                target_dims = list(data = c(leadtime_dim, time_dim, memb_dim), 
                                   params = c(time_dim, leadtime_dim, 'coef')),
                fun = .standardization, 
                coef = coef, 
                leadtime_dim = leadtime_dim, 
                time_dim = time_dim,
                memb_dim = memb_dim,
                handle_infinity = handle_infinity,
                cross_validation = cross_validation,
                method = method, distribution = distribution, fit = fit, 
                ref_period = ref_period, param_error = param_error, 
                na.rm = na.rm, 
                output_dims = list(data_spei = c(leadtime_dim, time_dim, memb_dim), 
                                   params = c(time_dim, leadtime_dim, 'coef')), 
                ncores = ncores)

  return(spei) # spei is a list of data_spei and params
}

.standardization <- function(data, params, coef, leadtime_dim = 'time', 
                             time_dim = 'syear', memb_dim = 'ensemble', 
                             ref_period = NULL, handle_infinity = FALSE, 
                             cross_validation = FALSE, param_error = -9999, 
                             method = 'parametric', distribution = 'log-Logistic',
                             fit = 'ub-pwm', na.rm = FALSE) {
                        
  # data: [leadtime_dim, time_dim, memb_dim]
  # params: [time_dim, leadtime_dim, 'coef']

  # maximum number of parameters needed to define any of the considered distributions
  ncoef <- length(coef) 
  nleadtime <- as.numeric(dim(data)[leadtime_dim])
  ntime <- as.numeric(dim(data)[time_dim])
  nmemb <- as.numeric(dim(data)[memb_dim])

  if (cross_validation) {
    params_result <- array(data = NA, dim = c(ntime, nleadtime, ncoef))
  } else {
    params_result <- array(data = NA, dim = c(1, nleadtime, ncoef))
  }
  names(dim(params_result)) <- c(time_dim, leadtime_dim, 'coef')

  if (all(is.na(data))) {
    spei_mod <- array(NA, dim(data))
    # if the data [time, sdate, memb] has no variability it will raise an error 
    # further down, so we assign a value to the result and skip the step
  } else if (any(is.na(data)) && !na.rm) {
    spei_mod <- array(NA, dim(data))
  } else if (var(data, na.rm = T) == 0) {                              
    spei_mod <- array(param_error, dim(data))
  } else {
    if (is.null(ref_period)) {
      ref.start <- NULL
      ref.end <- NULL
    } else {
      ref.start <- ref_period[[1]]
      ref.end <- ref_period[[2]]
    }

    spei_mod <- array(data = NA, dim = c(nleadtime, ntime, nmemb))
    names(dim(spei_mod)) <- c(leadtime_dim, time_dim, memb_dim)

    for (ff in 1:nleadtime) {
      data_subset <- ClimProjDiags::Subset(data, along = leadtime_dim, 
                                           indices = ff, drop = 'selected')
      
      params_tmp <- if (all(is.na(params))) {NULL} else {params[, ff, ]}


      spei_data <- .std(data = data_subset, coef = coef, 
                        ntime = ntime, nmemb = nmemb, 
                        method = method, distribution = distribution, 
                        fit = fit, na.rm = na.rm, 
                        ref.start = ref.start, ref.end = ref.end, 
                        params = params_tmp, 
                        handle_infinity = handle_infinity, 
                        cross_validation = cross_validation)
      spei_mod[ff, , ] <- spei_data[[1]]
      params_ff <- spei_data[[2]]
      # lengthen dimension coef of params_ff in case it doesn't match the 
      # corresponding dimension of parms_months
      if (!is.null(params_ff)) {
        if (length(params_ff) < ncoef) {
          params_ff <- append(params_ff, array(NA, dim = ncoef - length(params_ff)))
        }
        params_result[, ff, ] <- params_ff
      }
    }
  }
  return(list(spei = spei_mod, params = params_result))
}

.std <- function(data, coef, ntime, nmemb, method = 'parametric', 
                 distribution = 'log-Logistic', fit = 'ub-pwm', na.rm = FALSE, 
                 ref.start = NULL, ref.end = NULL, params = NULL, 
                 handle_infinity = FALSE, cross_validation = FALSE) {
  
  # data: [time_dim, memb_dim]
  # params: NULL or [(ntime), coef]
  if (method == 'non-parametric') { 
    bp <- matrix(0, length(data), 1)
    for (i in 1:length(data)) {
      bp[i,1] = sum(data[] <= data[i], na.rm = na.rm);  # Writes the rank of the data
    }
    std_index <- qnorm((bp - 0.44)/(length(data) + 0.12))
    dim(std_index) <- c(ntime, nmemb)
    # it won't return params to be used in exp_cor; also it is not using 
    # handle_infinity nor cross_validation
    params_result <- NULL
  } else {
    std_index <- array(NA, c(ntime, nmemb))
    # Select window if necessary
    if (!is.null(ref.start) && !is.null(ref.end)) {
      data.fit <- window(data, ref.start, ref.end)	
    } else {
      data.fit <- data
    }

    if (cross_validation) {  
      loop_years <- ntime
    } else {
      loop_years <- 1
    }

    params_result <- array(NA, dim = c(loop_years, length(coef)))
    colnames(params_result) <- names(coef)
    
    for (nsd in 1:loop_years) { # loop over years (in case of cross_validation)
      # Cumulative series (acu)
      if (cross_validation) {
        acu <- as.vector(data.fit[-nsd, ])
      } else {
        acu <- as.vector(data.fit)
      }

      acu.sorted <- sort.default(acu, method = "quick")
      # remove NAs (no need if(na.rm) because if there are NA and na.rm = F 
      # we don't get to this point)
      acu.sorted <- acu.sorted[!is.na(acu.sorted)] 
      # else all acu was NA and we don't need to continue with this case   
      
      if (length(acu.sorted) != 0) { 
        acu_sd <- sd(acu.sorted)
        if (!is.na(acu_sd)) {
          if (acu_sd != 0) {
            if (distribution != "log-Logistic") {
              pze <- sum(acu == 0) / length(acu)
              acu.sorted = acu.sorted[acu.sorted > 0]
            }
            if (!is.null(params)) {
              f_params <- as.vector(params)
              params_result[nsd, ] <- f_params
            } else {
              # else coef will be NA
              if (length(acu.sorted) >= 4) { 
                # Calculate probability weighted moments based on fit with lmomco or TLMoments
                pwm = switch(fit,
                             "pp-pwm" = pwm.pp(acu.sorted, -0.35, 0, nmom = 3),
                             pwm.ub(acu.sorted, nmom = 3)
                             # TLMoments::PWM(acu.sorted, order = 0:2)
                             )

                # Check L-moments validity
                lmom <- pwm2lmom(pwm)          
                if (!(!are.lmom.valid(lmom) || anyNA(lmom[[1]]) || any(is.nan(lmom[[1]])))) {
                  # lmom fortran functions need specific inputs L1, L2, T3
                  # this is handled by lmomco internally with lmorph
                  fortran_vec = c(lmom$lambdas[1:2], lmom$ratios[3])
                  # Calculate parameters based on distribution with lmom then lmomco
                  f_params = switch(distribution,
                                    "log-Logistic" = tryCatch(lmom::pelglo(fortran_vec), 
                                                              error = function(e){parglo(lmom)$para}),
                                    "Gamma" = tryCatch(lmom::pelgam(fortran_vec), 
                                                       error = function(e){pargam(lmom)$para}),
                                    "PearsonIII" = tryCatch(lmom::pelpe3(fortran_vec), 
                                                            error = function(e){parpe3(lmom)$para}))
                  # Adjust if user chose log-Logistic and max-lik
                  if (distribution == 'log-Logistic' && fit == 'max-lik') {
                    f_params = parglo.maxlik(acu.sorted, f_params)$para
                  }
                  params_result[nsd, ] <- f_params
                } # end for the case the L-moments are not valid (std_index will be NA)
              } # end for case there are not enough values to estimate the parameters (std_index will be NA)
            } # end estimation of f_param
            # Calculate cdf based on distribution with lmom
            if (all(is.na(params_result[nsd,]))) {
              cdf_res <- NA
            } else {
              f_params <- params_result[nsd,]
              f_params <- f_params[which(!is.na(f_params))]
              cdf_res = switch(distribution,
                               "log-Logistic" = lmom::cdfglo(data, f_params),
                               "Gamma" = lmom::cdfgam(data, f_params),
                               "PearsonIII" = lmom::cdfpe3(data, f_params))
            }

            std_index_cv <- array(qnorm(cdf_res), dim = c(ntime, nmemb))
    
            # Adjust if user chose Gamma or PearsonIII - Not tested: For future development
            # if (distribution != 'log-Logistic') { 
            #   std_index[ff,s] = qnorm(pze + (1-pze)*pnorm(std_index[ff,s])) # ff doesn't exist at this point
            # }
            if (cross_validation) {
              std_index[nsd, ] <- std_index_cv[nsd, ]  
            } else {
              std_index <- std_index_cv
            }
          }
        } # end if for the case there is no variability        
      } # end if for the case all NA in acu
    } # next year (in case of cross_validation or all done if cross_validation == F)

    if (handle_infinity) { # could also use "param_error" ?; we are giving it the min/max value of the grid point
      std_index[is.infinite(std_index) & std_index < 0] <- min(std_index[!is.infinite(std_index)])
      std_index[is.infinite(std_index) & std_index > 0] <- max(std_index[!is.infinite(std_index)]) 
    }
    # f_params will be params only if cross_validation is FALSE
    # (otherwise there will be one f_params per year;
    # but the output params will be read only in the case that
    # it is called with cross_validation FALSE)
  }
  return(list(std_index, params_result)) 
}
