% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PeriodSPEI.R
\name{CST_PeriodSPEI}
\alias{CST_PeriodSPEI}
\title{Compute the Standardised Precipitation-Evapotranspiration Index}
\usage{
CST_PeriodSPEI(
  exp,
  exp_cor = NULL,
  time_dim = "syear",
  leadtime_dim = "time",
  memb_dim = "ensemble",
  lat_dim = "latitude",
  accum = 1,
  ref_period = NULL,
  params = NULL,
  pet_exp = NULL,
  pet_expcor = NULL,
  standardization = TRUE,
  cross_validation = FALSE,
  pet_method = "hargreaves",
  method = "parametric",
  distribution = "log-Logistic",
  fit = "ub-pwm",
  param_error = -9999,
  handle_infinity = FALSE,
  na.rm = FALSE,
  ncores = NULL
)
}
\arguments{
\item{exp}{A named list with the needed \code{s2dv_cube} objects containing 
the seasonal forecast experiment in the data element for each variable. 
Specific variables are needed for each method used in computing the 
Potential Evapotranspiration. See parameter 'pet_method'. The accepted 
variables are: 'tasmin' and 'tasmax', required for methods 'hargreaves' and 
'hargreaves_modified' and 'tas', required for method 'thornthwaite'. 
Variable 'prlr' is always needed. The units for temperature variables
('tasmin', 'tasmax' and 'tas') need to be in Celcius degrees; the units for 
precipitation ('prlr') need to be in mm/month.}

\item{exp_cor}{A named list with the needed \code{s2dv_cube} objects for each 
variable in which the quantile PeriodSPEI should be applied. If it is not 
specified, the PeriodSPEI is calculated from object 'exp'.}

\item{time_dim}{A character string indicating the name of the temporal 
dimension. By default, it is set to 'syear'.}

\item{leadtime_dim}{A character string indicating the name of the temporal 
dimension. By default, it is set to 'time'.}

\item{memb_dim}{A character string indicating the name of the dimension in 
which the ensemble members are stored. When set it to NULL, threshold is 
computed for individual members.}

\item{lat_dim}{A character string indicating the name of the latitudinal 
dimension. By default it is set by 'latitude'.}

\item{accum}{An integer value indicating the number of months for the 
accumulation for each variable.}

\item{ref_period}{A list with two numeric values with the starting and end 
points of the reference period used for computing the index. The default 
value is NULL indicating that the first and end values in data will be 
used as starting and end points.}

\item{params}{A multidimensional array with named dimensions for computing the 
SPEI. This option overrides computation of fitting parameters. It needs 
to be of same leadtime and time dimensions of 'exp' and a dimension named 
'coef' with the length of the coefficients needed for the used distribution 
(for 'Gamma' coef dimension is of lenght 2, for 'log-Logistic' or 
'PearsonIII' is of length 3). If cross_validation is FALSE, the leadtime 
dimension must be of length 1.}

\item{pet_exp}{A multidimensional array containing the Potential
EvapoTranspiration data of 'exp'. It must have the same dimensions of the 
variable 'prlr' of 'exp'. If it is NULL it is calculated using the provided 
variables with specified 'pet_method'. It is NULL by default.}

\item{pet_expcor}{A multidimensional array containing the Potential
EvapoTranspiration data of 'exp_cor'. It must have the same dimensions of 
the variable 'prlr' of 'exp_cor'. If it is NULL it is calculated using the 
provided variables with specified 'pet_method'. It is NULL by default.}

\item{standardization}{A logical value indicating wether the standardization 
is computed.}

\item{cross_validation}{A logical value indicating if cross validation is 
done (TRUE), or not (FALSE). It is FALSE by default.}

\item{pet_method}{A character string indicating the method used to compute 
the potential evapotranspiration. The accepted methods are:
'hargreaves' and 'hargreaves_modified', that require the data to have 
variables tasmin and tasmax; and 'thornthwaite', that requires variable 
'tas'.}

\item{method}{A character string indicating the standardization method used. 
If can be: 'parametric' or 'non-parametric'. It is set to 'parametric' by 
default.}

\item{distribution}{A character string indicating the name of the distribution 
function to be used for computing the SPEI. The accepted names are: 
'log-Logistic', 'Gamma' or 'PearsonIII'. It is set to 'log-Logistic' by 
default.}

\item{fit}{A character string indicating the name of the method used for 
computing the distribution function parameters The accepteed names are: 
'ub-pwm', 'pp-pwm' and 'max-lik'. It is set to 'ub-pwm' by default.}

\item{param_error}{A numeric value with the error accepted.}

\item{handle_infinity}{A logical value wether to return Infinite values (TRUE)
or not (FALSE).}

\item{na.rm}{A logical value indicating whether NA values should be removed 
from data. It is FALSE by default.}

\item{ncores}{An integer value indicating the number of cores to use in 
parallel computation.}
}
\value{
A list with elements:
\itemize{
 \item{'exp', if 'standarization' is TRUE an 's2dv_cube' conaining the 'SPEI' 
       in element data from 'exp' array with the same dimensions as 'exp'.
       If it is FALSE, it is an array with the accumulated values of PET 
       minus 'prlr' data.}
 \item{'exp_cor', if 'standarization' is TRUE and if 'exp_cor' is not 
       NULL. It is an 's2dv_cube' with the SPEI data from 'exp_cor' in 
       element 'data'. If 'standarization' is FALSE, only the accumulated 
       values of PET minus 'prlr' is returned.}
 \item{'params', returned if 'standarization' is TRUE, it contains the 
       parameters used for the standarization of 'exp' that are used for 
       computing the 'SPEI' for 'exp_cor'.}
}
}
\description{
Calculation of the Standardised Precipitation-Evapotranspiration Index (SPEI) 
that is a multiscalar drought index based on climatic data. It can be used for 
determining the onset, duration and magnitude of drought conditions with 
respect to normal conditions in a variety of natural and managed systems such 
as crops, ecosystems, rivers, water resources, etc. The SPI is calculated 
using monthly (or weekly) precipitation as the input data. The SPEI uses the 
monthly (or weekly) difference between precipitation and pet. This represents 
a simple climatic water balance which is calculated at different time scales 
to obtain the SPEI.
}
\examples{
dims <- c(var = 1, sday = 1, sweek = 1, syear = 6, time = 3, 
         latitude = 2, longitude = 1, ensemble = 25)
dimscor <- c(var = 1, sday = 1, sweek = 1, syear = 1, time = 3, 
            latitude = 2, longitude = 1, ensemble = 15)

dates_exp <- as.POSIXct(c(paste0(2010:2015, "-08-16"), 
                         paste0(2010:2015, "-09-15"), 
                         paste0(2010:2015, "-10-16")), 'UTC')
dim(dates_exp) <- c(sday = 1, sweek = 1, syear = 6, time = 3)
dates_expcor <- as.POSIXct(c(paste0(2020, "-08-16"), paste0(2020, "-09-15"), 
                            paste0(2020, "-10-16")), 'UTC')
dim(dates_expcor) <- c(sday = 1, sweek = 1, syear = 1, time = 3)
lat <- c(40,40.1)

exp_tasmax <- array(rnorm(900, 27.73, 5.26), dim = dims)
exp_tasmin <- array(rnorm(900, 14.83, 3.86), dim = dims)
exp_prlr <- array(rnorm(900, 21.19, 25.64), dim = dims)

expcor_tasmax <- array(rnorm(540, 29.03, 5.67), dim = dimscor)
expcor_tasmin <- array(rnorm(540, 15.70, 4.40), dim = dimscor)
expcor_prlr <- array(rnorm(540, 15.62, 21.38), dim = dimscor)

exp <- list('tasmax' = exp_tasmax, 'tasmin' = exp_tasmin, 'prlr' = exp_prlr)
exp_cor <- list('tasmax' = expcor_tasmax, 'tasmin' = expcor_tasmin, 
               'prlr' = expcor_prlr)

exp <- lapply(exp, CSTools::s2dv_cube, coords = list(latitude = lat), 
             Dates = dates_exp)
exp_cor <- lapply(exp_cor, CSTools::s2dv_cube, coords = list(latitude = lat), 
                 Dates = dates_expcor)

res <- CST_PeriodSPEI(exp = exp, exp_cor = exp_cor)

}
